@@ -5,6 +5,8 @@ import yaml
 import string
 from textwrap import dedent
 from pathlib import Path, PurePath
+from copy import copy
+
 from regexapp.exceptions import EscapePatternError
 from regexapp.exceptions import PatternReferenceError
 from regexapp.exceptions import TextPatternError
@@ -384,6 +386,11 @@ class TextPattern(str):
     Methods
     -------
     TextPattern.get_pattern(text) -> str
+    lstrip(chars=None) -> TextPattern
+    rstrip(chars=None) -> TextPattern
+    strip(chars=None) -> TextPattern
+    add(other, as_is=True) -> TextPattern
+    concatenate(*other, as_is=True) -> TextPattern
 
     Raises
     ------
@@ -405,6 +412,18 @@ class TextPattern(str):
         self.text = text
         self.as_is = as_is
 
+    def __add__(self, other):
+        result = super().__add__(other)
+        result_pat = TextPattern(result, as_is=True)
+        return result_pat
+
+    def __radd__(self, other):
+        if isinstance(other, TextPattern):
+            return other.__add__(self)
+        else:
+            other_pat = TextPattern(other, as_is=True)
+            return other_pat.__add__(self)
+
     @property
     def is_empty(self):
         if self == '':
@@ -516,6 +535,52 @@ class TextPattern(str):
         pattern = TextPattern(new_text)
         return pattern
 
+    def add(self, other, as_is=True):
+        """return a concatenated TextPattern.
+
+        Parameters
+        ----------
+        other (str, TextElement): other
+        as_is (bool): a flag to keep adding other AS-IS condition.
+
+        Returns
+        -------
+        TextPattern: a concatenated TextPattern instance.
+        """
+        if isinstance(other, TextPattern):
+            result = self + other
+        else:
+            if isinstance(other, (list, tuple)):
+                result = self
+                for item in other:
+                    if isinstance(item, TextPattern):
+                        result = result + item
+                    else:
+                        item_pat = TextPattern(str(item), as_is=as_is)
+                        result = result + item_pat
+            else:
+                other_pat = TextPattern(str(other), as_is=as_is)
+                result = self + other_pat
+
+        return result
+
+    def concatenate(self, *other, as_is=True):
+        """return a concatenated TextPattern.
+
+        Parameters
+        ----------
+        other (tuple): other
+        as_is (bool): a flag to keep adding other AS-IS condition.
+
+        Returns
+        -------
+        TextPattern: a concatenated TextPattern instance.
+        """
+        result = self
+        for item in other:
+            result = result.add(item, as_is=as_is)
+        return result
+
 
 class ElementPattern(str):
     """Use to convert element data to regex pattern
@@ -549,14 +614,36 @@ class ElementPattern(str):
     ElementPattern.add_start_of_string(pattern, head='') -> str
     ElementPattern.add_end_of_string(pattern, tail='') -> str
     ElementPattern.add_repetition(lst, repetition='') -> list
+    ElementPattern.add_occurrence(lst, occurrence='') -> list
+    ElementPattern.add_case_occurrence(lst, first, last, is_phrase) -> bool
+    ElementPattern.is_singular_pattern(pattern) -> bool
+    remove_head_of_string() -> ElementPattern
+    remove_tail_of_string() -> ElementPattern
 
     Raises
     ------
     ElementPatternError: raise an exception if pattern is invalid.
     """
+    # patterns
+    word_bound_pattern = r'word_bound(_left|_right|_raw)?$'
+    head_pattern = r'head(_raw|((_just)?_(whitespace|ws|space)(_plus)?))?$'
+    tail_pattern = r'tail(_raw|((_just)?_(whitespace|ws|space)(_plus)?))?$'
+    repetition_pattern = r'repetition_\d*(_\d*)?$'
+    occurrence_pattern = r'({})(?P<is_phrase>_phrase)?_occurrences?$'.format(
+        '|'.join([
+            r'((?P<fda>\d+)_or_(?P<lda>\d+))',
+            r'((?P<fdb>\d+)_or_(?P<ldb>more))',
+            r'(at_(?P<fdc>least|most)_(?P<ldc>\d+))',
+            r'(?P<fdd>\d+)'
+        ])
+    )
+    meta_data_pattern = r'^meta_data_\w+'
+
     def __new__(cls, text, as_is=False):
         cls._variable = VarCls()
         cls._or_empty = False
+        cls._prepended_pattern = ''
+        cls._appended_pattern = ''
         data = str(text)
 
         if as_is:
@@ -573,10 +660,14 @@ class ElementPattern(str):
         self.as_is = as_is
         self.variable = self._variable
         self.or_empty = self._or_empty
+        self.prepended_pattern = self._prepended_pattern
+        self.appended_pattern = self._appended_pattern
 
         # clear class variable after initialization
         self._variable = VarCls()
         self._or_empty = False
+        self._prepended_pattern = ''
+        self._appended_pattern = ''
 
     @classmethod
     def get_pattern(cls, text):
@@ -682,37 +773,35 @@ class ElementPattern(str):
         tail = ''
         is_repeated = False
         is_occurrence = False
-        occurrence_pat = (r'((\d+_or_\d+)|(\d+_or_more)|(at_(least|most)_\d+)'
-                          r'|\d+)(_phrase)?_occurrences?$')
 
         for arg in arguments:
             match = re.match(vpat, arg, flags=re.I)
             if match:
                 name = match.group('name') if not name else name
-            elif re.match('word_bound(_left|_right|_raw)?$', arg):
+            elif re.match(cls.word_bound_pattern, arg):
                 if arg == 'word_bound_raw':
                     'word_bound' not in lst and lst.append('word_bound')
                 else:
                     word_bound = arg
-            elif re.match('head(_raw|(_(ws|space)(_plus)?))?$', arg):
+            elif re.match(cls.head_pattern, arg):
                 if arg == 'head_raw':
                     'head' not in lst and lst.append('head')
                 else:
                     head = arg
-            elif re.match('tail(_raw|(_(ws|space)(_plus)?))?$', arg):
+            elif re.match(cls.tail_pattern, arg):
                 if arg == 'tail_raw':
                     'tail' not in lst and lst.append('tail')
                 else:
                     tail = arg
-            elif re.match(r'repetition_\d*(_\d*)?$', arg):
+            elif re.match(cls.repetition_pattern, arg):
                 if not is_repeated or not is_occurrence:
                     lst = cls.add_repetition(lst, repetition=arg)
                     is_repeated = True
-            elif re.match(occurrence_pat, arg):
+            elif re.match(cls.occurrence_pattern, arg):
                 if not is_repeated or not is_occurrence:
                     lst = cls.add_occurrence(lst, occurrence=arg)
                     is_occurrence = True
-            elif re.match(r'^meta_data_\w+', arg):
+            elif re.match(cls.meta_data_pattern, arg):
                 if arg == 'meta_data_raw':
                     'meta_data' not in lst and lst.append('meta_data')
                 else:
@@ -787,37 +876,35 @@ class ElementPattern(str):
         tail = ''
         is_repeated = False
         is_occurrence = False
-        occurrence_pat = (r'((\d+_or_\d+)|(\d+_or_more)|(at_(least|most)_\d+)'
-                          r'|\d+)(_phrase)?_occurrences?$')
 
         for arg in arguments:
             match = re.match(vpat, arg, flags=re.I)
             if match:
                 name = match.group('name') if not name else name
-            elif re.match('word_bound(_left|_right|_raw)?$', arg):
+            elif re.match(cls.word_bound_pattern, arg):
                 if arg == 'word_bound_raw':
                     'word_bound' not in lst and lst.append('word_bound')
                 else:
                     word_bound = arg
-            elif re.match('head(_raw|(_(ws|space)(_plus)?))?$', arg):
+            elif re.match(cls.head_pattern, arg):
                 if arg == 'head_raw':
                     'head' not in lst and lst.append('head')
                 else:
                     head = arg
-            elif re.match('tail(_raw|(_(ws|space)(_plus)?))?$', arg):
+            elif re.match(cls.tail_pattern, arg):
                 if arg == 'tail_raw':
                     'tail' not in lst and lst.append('tail')
                 else:
                     tail = arg
-            elif re.match(r'repetition_\d*(_\d*)?$', arg):
+            elif re.match(cls.repetition_pattern, arg):
                 if not is_repeated or not is_occurrence:
                     lst = cls.add_repetition(lst, repetition=arg)
                     is_repeated = True
-            elif re.match(occurrence_pat, arg):
+            elif re.match(cls.occurrence_pattern, arg):
                 if not is_repeated or not is_occurrence:
                     lst = cls.add_occurrence(lst, occurrence=arg)
                     is_occurrence = True
-            elif re.match(r'^meta_data_\w+', arg):
+            elif re.match(cls.meta_data_pattern, arg):
                 if arg == 'meta_data_raw':
                     'meta_data' not in lst and lst.append('meta_data')
                 else:
@@ -899,22 +986,22 @@ class ElementPattern(str):
             match = re.match(vpat, arg, flags=re.I)
             if match or arg.startswith('format'):
                 continue
-            elif re.match('word_bound(_left|_right|_raw)?$', arg):
+            elif re.match(cls.word_bound_pattern, arg):
                 if arg == 'word_bound_raw':
                     'word_bound' not in lst and lst.append('word_bound')
                 else:
                     word_bound = arg
-            elif re.match('head(_raw|(_(ws|space)(_plus)?))?$', arg):
+            elif re.match(cls.head_pattern, arg):
                 if arg == 'head_raw':
                     'head' not in lst and lst.append('head')
                 else:
                     head = arg
-            elif re.match('tail(_raw|(_(ws|space)(_plus)?))?$', arg):
+            elif re.match(cls.tail_pattern, arg):
                 if arg == 'tail_raw':
                     'tail' not in lst and lst.append('tail')
                 else:
                     tail = arg
-            elif re.match(r'^meta_data_\w+', arg):
+            elif re.match(cls.meta_data_pattern, arg):
                 if arg == 'meta_data_raw':
                     'meta_data' not in lst and lst.append('meta_data')
                 else:
@@ -976,22 +1063,22 @@ class ElementPattern(str):
             match = re.match(vpat, arg, flags=re.I)
             if match:
                 name = match.group('name') if not name else name
-            elif re.match('word_bound(_left|_right|_raw)?$', arg):
+            elif re.match(cls.word_bound_pattern, arg):
                 if arg == 'word_bound_raw':
                     'word_bound' not in lst and lst.append('word_bound')
                 else:
                     word_bound = arg
-            elif re.match('head(_raw|(_(ws|space)(_plus)?))?$', arg):
+            elif re.match(cls.head_pattern, arg):
                 if arg == 'head_raw':
                     'head' not in lst and lst.append('head')
                 else:
                     head = arg
-            elif re.match('tail(_raw|(_(ws|space)(_plus)?))?$', arg):
+            elif re.match(cls.tail_pattern, arg):
                 if arg == 'tail_raw':
                     'tail' not in lst and lst.append('tail')
                 else:
                     tail = arg
-            elif re.match(r'^meta_data_\w+', arg):
+            elif re.match(cls.meta_data_pattern, arg):
                 if arg == 'meta_data_raw':
                     'meta_data' not in lst and lst.append('meta_data')
                 else:
@@ -1053,22 +1140,22 @@ class ElementPattern(str):
             match = re.match(vpat, arg, flags=re.I)
             if match:
                 name = match.group('name') if not name else name
-            elif re.match('word_bound(_left|_right|_raw)?$', arg):
+            elif re.match(cls.word_bound_pattern, arg):
                 if arg == 'word_bound_raw':
                     'word_bound' not in lst and lst.append('word_bound')
                 else:
                     word_bound = arg
-            elif re.match('head(_raw|(_(ws|space)(_plus)?))?$', arg):
+            elif re.match(cls.head_pattern, arg):
                 if arg == 'head_raw':
                     'head' not in lst and lst.append('head')
                 else:
                     head = arg
-            elif re.match('tail(_raw|(_(ws|space)(_plus)?))?$', arg):
+            elif re.match(cls.tail_pattern, arg):
                 if arg == 'tail_raw':
                     'tail' not in lst and lst.append('tail')
                 else:
                     tail = arg
-            elif re.match(r'^meta_data_\w+', arg):
+            elif re.match(cls.meta_data_pattern, arg):
                 if arg == 'meta_data_raw':
                     'meta_data' not in lst and lst.append('meta_data')
                 else:
@@ -1300,16 +1387,51 @@ class ElementPattern(str):
             case3, case4 = r'^ *', r'^ +'
             case5 = r'^'
 
+            case6, case7 = r'\s*', r'\s+'
+            case8, case9 = r' *', r' +'
+
+            case10, case11 = r'^\s*', r'^\s+'
+            case12, case13 = r'\s*', r'\s+'
+
             if head == 'head_ws' and not pattern.startswith(case1):
                 new_pattern = '{}{}'.format(case1, pattern)
+                cls._prepended_pattern = case1
             elif head == 'head_ws_plus' and not pattern.startswith(case2):
                 new_pattern = '{}{}'.format(case2, pattern)
+                cls._prepended_pattern = case2
             elif head == 'head_space' and not pattern.startswith(case3):
                 new_pattern = '{}{}'.format(case3, pattern)
+                cls._prepended_pattern = case3
             elif head == 'head_space_plus' and not pattern.startswith(case4):
                 new_pattern = '{}{}'.format(case4, pattern)
+                cls._prepended_pattern = case4
             elif head == 'head' and not pattern.startswith(case5):
                 new_pattern = '{}{}'.format(case5, pattern)
+                cls._prepended_pattern = case5
+            elif head == 'head_just_ws' and not pattern.startswith(case6):
+                new_pattern = '{}{}'.format(case6, pattern)
+                cls._prepended_pattern = case6
+            elif head == 'head_just_ws_plus' and not pattern.startswith(case7):
+                new_pattern = '{}{}'.format(case7, pattern)
+                cls._prepended_pattern = case7
+            elif head == 'head_just_space' and not pattern.startswith(case8):
+                new_pattern = '{}{}'.format(case8, pattern)
+                cls._prepended_pattern = case8
+            elif head == 'head_just_space_plus' and not pattern.startswith(case9):
+                new_pattern = '{}{}'.format(case9, pattern)
+                cls._prepended_pattern = case9
+            elif head == 'head_whitespace' and not pattern.startswith(case10):
+                new_pattern = '{}{}'.format(case10, pattern)
+                cls._prepended_pattern = case10
+            elif head == 'head_whitespace_plus' and not pattern.startswith(case11):
+                new_pattern = '{}{}'.format(case11, pattern)
+                cls._prepended_pattern = case11
+            elif head == 'head_just_whitespace' and not pattern.startswith(case12):
+                new_pattern = '{}{}'.format(case12, pattern)
+                cls._prepended_pattern = case12
+            elif head == 'head_just_whitespace_plus' and not pattern.startswith(case13):
+                new_pattern = '{}{}'.format(case13, pattern)
+                cls._prepended_pattern = case13
             else:
                 new_pattern = pattern
             return new_pattern
@@ -1332,16 +1454,52 @@ class ElementPattern(str):
             case1, case2 = r'\s*$', r'\s+$'
             case3, case4 = r' *$', r' +$'
             case5 = r'$'
+
+            case6, case7 = r'\s*', r'\s+'
+            case8, case9 = r' *', r' +'
+
+            case10, case11 = r'\s*$', r'\s+$'
+            case12, case13 = r'\s*', r'\s+'
+
             if tail == 'tail_ws' and not pattern.endswith(case1):
                 new_pattern = '{}{}'.format(pattern, case1)
+                cls._appended_pattern = case1
             elif tail == 'tail_ws_plus' and not pattern.endswith(case2):
                 new_pattern = '{}{}'.format(pattern, case2)
+                cls._appended_pattern = case2
             elif tail == 'tail_space' and not pattern.endswith(case3):
                 new_pattern = '{}{}'.format(pattern, case3)
+                cls._appended_pattern = case3
             elif tail == 'tail_space_plus' and not pattern.endswith(case4):
                 new_pattern = '{}{}'.format(pattern, case4)
+                cls._appended_pattern = case4
             elif tail == 'tail' and not pattern.endswith(case5):
                 new_pattern = '{}{}'.format(pattern, case5)
+                cls._appended_pattern = case5
+            elif tail == 'tail_just_ws' and not pattern.startswith(case6):
+                new_pattern = '{}{}'.format(pattern, case6)
+                cls._appended_pattern = case6
+            elif tail == 'tail_just_ws_plus' and not pattern.startswith(case7):
+                new_pattern = '{}{}'.format(pattern, case7)
+                cls._appended_pattern = case7
+            elif tail == 'tail_just_space' and not pattern.startswith(case8):
+                new_pattern = '{}{}'.format(pattern, case8)
+                cls._appended_pattern = case8
+            elif tail == 'tail_just_space_plus' and not pattern.startswith(case9):
+                new_pattern = '{}{}'.format(pattern, case9)
+                cls._appended_pattern = case9
+            elif tail == 'tail_whitespace' and not pattern.startswith(case10):
+                new_pattern = '{}{}'.format(pattern, case10)
+                cls._appended_pattern = case10
+            elif tail == 'tail_whitespace_plus' and not pattern.startswith(case11):
+                new_pattern = '{}{}'.format(pattern, case11)
+                cls._appended_pattern = case11
+            elif tail == 'tail_just_whitespace' and not pattern.startswith(case12):
+                new_pattern = '{}{}'.format(pattern, case12)
+                cls._appended_pattern = case12
+            elif tail == 'tail_just_whitespace_plus' and not pattern.startswith(case13):
+                new_pattern = '{}{}'.format(pattern, case13)
+                cls._appended_pattern = case13
             else:
                 new_pattern = pattern
             return new_pattern
@@ -1393,17 +1551,8 @@ class ElementPattern(str):
         if not occurrence:
             return lst
 
-        occurrence_pat = r'({})(?P<is_phrase>_phrase)?_occurrences?$'.format(
-            '|'.join([
-                r'((?P<fda>\d+)_or_(?P<lda>\d+))',
-                r'((?P<fdb>\d+)_or_(?P<ldb>more))',
-                r'(at_(?P<fdc>least|most)_(?P<ldc>\d+))',
-                r'(?P<fdd>\d+)'
-            ])
-        )
-
         new_lst = lst[:]
-        m = re.match(occurrence_pat, occurrence)
+        m = re.match(cls.occurrence_pattern, occurrence)
         is_phrase = bool(m.group('is_phrase'))
 
         fda, lda = m.group('fda') or '', m.group('lda') or ''
@@ -1490,6 +1639,46 @@ class ElementPattern(str):
         is_char_set &= pattern.count(last) == 1 and last == right_bracket
         return is_singular or is_escape or is_char_set
 
+    def remove_head_of_string(self):
+        """remove a start of string pattern i.e ^ or ^\\s* or ^\\s+ or ^ * or ^ +
+
+        Returns
+        -------
+        ElementPattern: new ElementPattern
+        """
+        if self.prepended_pattern and self.startswith('^'):
+            pattern = str(self)[len(self.prepended_pattern):]
+            new_instance = ElementPattern(pattern, as_is=True)
+            new_instance.as_is = False
+            new_instance.variable = copy(self.variable)
+            new_instance.or_empty = self.or_empty
+            new_instance.prepended_pattern = ''
+            new_instance.appended_pattern = self.appended_pattern
+        else:
+            new_instance = copy(self)
+
+        return new_instance
+
+    def remove_tail_of_string(self):
+        """remove an end of string pattern i.e $ or \\s*$ or \\s+$ or  *$ or  +$
+
+        Returns
+        -------
+        ElementPattern: new ElementPattern
+        """
+        if self.appended_pattern and self.endswith('$'):
+            pattern = str(self)[:-len(self.appended_pattern)]
+            new_instance = ElementPattern(pattern, as_is=True)
+            new_instance.as_is = False
+            new_instance.variable = copy(self.variable)
+            new_instance.or_empty = self.or_empty
+            new_instance.prepended_pattern = self.prepended_pattern
+            new_instance.appended_pattern = ''
+        else:
+            new_instance = copy(self)
+
+        return new_instance
+
 
 class LinePattern(str):
     """Use to convert a line text to regex pattern
@@ -1641,14 +1830,64 @@ class LinePattern(str):
         if len(lst) < 2:
             return
 
+        total = len(lst)
+
         ws_pat = r'\s*'
-        for index, item in enumerate(lst[:-1]):
-            next_item = lst[index+1]
-            is_item_text_pat = isinstance(item, TextPattern)
-            is_next_item_elm_pat = isinstance(next_item, ElementPattern)
-            if is_item_text_pat and is_next_item_elm_pat:
-                if item.is_whitespace and next_item.or_empty:
-                    lst[index] = ws_pat
+        for index, item in enumerate(lst[1:], 1):
+            prev_item = lst[index-1]
+            is_prev_item_text_pat = isinstance(prev_item, (TextPattern, str))
+            is_item_elm_pat = isinstance(item, ElementPattern)
+            if is_prev_item_text_pat and is_item_elm_pat:
+                if item.or_empty:
+                    if prev_item.endswith(' '):
+                        lst[index-1] = prev_item.rstrip() + ws_pat
+                    elif prev_item.endswith(r'\s'):
+                        lst[index-1] = prev_item + '*'
+                    elif index == total - 1:
+                        if prev_item.endswith(' +'):
+                            lst[index-1] = prev_item[:-2] + ws_pat
+                        elif prev_item.endswith(r'\s+'):
+                            lst[index-1] = prev_item[:-3] + ws_pat
+
+        index = len(lst) - 1
+        is_stopped = False
+        while index > 0 and not is_stopped:
+            prev_item, item = lst[index-1], lst[index]
+            is_prev_item_text_pat = isinstance(prev_item, (TextPattern, str))
+            is_item_elm_pat = isinstance(item, ElementPattern)
+            if is_prev_item_text_pat and is_item_elm_pat:
+                if item.or_empty:
+                    if prev_item.endswith(' '):
+                        lst[index - 1] = prev_item.rstrip() + ws_pat
+                    elif prev_item.endswith(r'\s'):
+                        lst[index - 1] = prev_item + '*'
+                    elif prev_item.endswith(' +'):
+                        lst[index - 1] = prev_item[:-2] + ws_pat
+                    elif prev_item.endswith(r'\s+'):
+                        lst[index - 1] = prev_item[:-3] + ws_pat
+            else:
+                is_stopped = True
+            index -= 2
+
+        index = len(lst) - 1
+        is_stopped = False
+        is_prev_containing_empty = False
+        while index > 0 and not is_stopped:
+            prev_item, item = lst[index-1], lst[index]
+            is_prev_item_elm_pat = isinstance(prev_item, ElementPattern)
+            is_item_text_pat = isinstance(item, (TextPattern, str))
+            if is_prev_item_elm_pat and is_item_text_pat:
+                if prev_item.or_empty:
+                    if item in [' ', ' +', r'\s', r'\s+']:
+                        lst[index] = ws_pat
+                    is_prev_containing_empty = True
+                else:
+                    if item in [' ', ' +', r'\s', r'\s+'] and is_prev_containing_empty:
+                        lst[index] = ws_pat
+                    is_prev_containing_empty = False
+            else:
+                is_stopped = True
+            index -= 2
 
     @classmethod
     def ensure_start_of_line_pattern(cls, lst):
@@ -1682,6 +1921,11 @@ class LinePattern(str):
                 else:
                     lst[1] = new_val
 
+        # clean up any invalid a start of string pattern
+        for index, node in enumerate(lst[1:], 1):
+            if isinstance(node, ElementPattern) and node.prepended_pattern:
+                lst[index] = node.remove_head_of_string()
+
     @classmethod
     def ensure_end_of_line_pattern(cls, lst):
         """Ensure an end pattern does not contain duplicate whitespace
@@ -1714,6 +1958,11 @@ class LinePattern(str):
                 else:
                     lst[-2] = new_val
 
+        # clean up any invalid a start of string pattern
+        for index, node in enumerate(lst[:-1]):
+            if isinstance(node, ElementPattern) and node.appended_pattern:
+                lst[index] = node.remove_tail_of_string()
+
     @classmethod
     def prepend_whitespace(cls, lst):
         """prepend whitespace pattern to list
